from __future__ import annotations

import sarus_data_spec.protobuf as sp

from sarus.dataspec_wrapper import DataSpecWrapper
from sarus.utils import register_ops, sarus_method, sarus_model

try:
    import sklearn.decomposition as decomposition
    from sklearn.decomposition import *
except ModuleNotFoundError:
    pass  # error message in sarus_data_spec.typing


class PCA(DataSpecWrapper[decomposition.PCA]):
    @sarus_model(sp.Scalar.Model.ModelClass.SK_PCA)
    def __init__(
        self,
        n_components=None,
        *,
        copy=True,
        whiten=False,
        svd_solver="auto",
        tol=0.0,
        iterated_power="auto",
        random_state=None,
        _dataspec=None,
    ):
        ...

    @sarus_method("sklearn.SK_FIT", inplace=True)
    def fit(self, X, y=None):
        ...


register_ops()



"""SDK classes and functions."""
import warnings

import nest_asyncio

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from sarus import (
        numpy,
        pandas,
        pandas_profiling,
        plotly,
        sklearn,
        std,
        xgboost,
    )

    from .sarus import Client, Dataset
    from .utils import eval, eval_policy, length

VERSION = "0.4.0"

__all__ = ["Dataset", "Client", "length", "eval", "eval_policy", "config"]


nest_asyncio.apply()



from sarus.dataspec_wrapper import DataSpecWrapper


class List(DataSpecWrapper[list]):
    ...


class Tuple(DataSpecWrapper[tuple]):
    ...


class Int(DataSpecWrapper[int]):
    ...


class Float(DataSpecWrapper[float]):
    ...


class String(DataSpecWrapper[str]):
    ...



import pandas as pd

from sarus.dataspec_wrapper import DataSpecWrapper


class Index(DataSpecWrapper[pd.Index]):
    ...


class RangeIndex(DataSpecWrapper[pd.RangeIndex]):
    ...


class DataFrameGroupBy(DataSpecWrapper[pd.core.groupby.DataFrameGroupBy]):
    ...



import io
import pickle as pkl
from http import HTTPStatus
from itertools import count
from time import sleep
from typing import List

import pyarrow as pa
import pyarrow.parquet as pq
import requests
import sarus_data_spec.typing as st
from sarus_data_spec.protobuf.utilities import dict_serialize

from sarus.typing import Client


def fetch_synthetic(client: Client, id: str) -> pa.Table:
    """Fetch synthetic data for a Dataset."""
    resp = client.session.get(
        f"{client.base_url}/synthetic_data/{id}",
        stream=True,
        params={
            "textual_categories": True,
            "rows_number": None,
        },
    )
    if resp.status_code > 200:
        raise Exception(
            f"Error while retrieving synthetic data. "
            f"Gateway answer was: \n{resp}"
        )

    synthetic_table = pq.ParquetFile(io.BytesIO(resp.content)).read()

    return synthetic_table


def post_graph(
    client: Client,
    target: st.DataSpec,
    dataspecs: List[st.DataSpec],
    transforms: List[st.Transform],
    attributes,
    execute: bool,
) -> requests.Response:
    """Post a DataSpec and all its ancestors and attributes."""
    # TODO Ask server what it already has
    payload = pkl.dumps(
        {
            "items": [
                {
                    "target": target.uuid(),
                    "dataspecs": [
                        dict_serialize(dataspec.protobuf())
                        for dataspec in dataspecs
                    ],
                    "transforms": [
                        dict_serialize(transform.protobuf())
                        for transform in transforms
                    ],
                    "attributes": [
                        dict_serialize(attribute.protobuf())
                        for attribute in attributes
                    ],
                }
            ]
        }
    )

    resp = client.session.post(
        url=f"{client.base_url}/v2/dataspecs",
        data=payload,
        params={"execute": "true" if execute else "false"},
    )

    return resp


def get_dataspec_status(
    client: Client,
    dataspecs: List[st.DataSpec],
) -> requests.Response:
    """Fetch information on a DataSpec status on the server."""
    return client.session.get(
        url=f"{client.base_url}/v2/dataspecs/status",
        params={"uuids": [ds.uuid() for ds in dataspecs]},
    )


def get_dataspec(
    client: Client,
    dataspec: st.DataSpec,
    max_attempts: int = 10,
    interval: int = 4,
) -> bytes:
    """Fetch a DataSpec value from the server."""
    resp = client.session.get(
        url=f"{client.base_url}/v2/dataspecs", params={"uuid": dataspec.uuid()}
    )
    if resp.status_code >= HTTPStatus.BAD_REQUEST:
        msg = resp.json()["message"]
        raise ValueError(f"Server - {msg}")

    attempts = count()
    while (
        resp.status_code == HTTPStatus.ACCEPTED
        and next(attempts) <= max_attempts
    ):
        sleep(interval)
        resp = client.session.get(
            url=f"{client.base_url}/v2/dataspecs",
            params={"uuid": dataspec.uuid()},
        )
        if resp.status_code >= HTTPStatus.BAD_REQUEST:
            msg = resp.json()["message"]
            raise ValueError(f"Server - {msg}")

    if resp.status_code != HTTPStatus.OK:
        msg = resp.json()["message"]
        raise ValueError(f"Server - {msg}")
    for data in resp.iter_content(chunk_size=1 << 10):
        yield data


def delete_dataspec(client: Client, uuid: str) -> requests.Response:
    """Delete a DataSpec from the server."""
    return client.session.delete(
        url=f"{client.base_url}/v2/dataspecs",
        params={"uuid": uuid},
    )



import inspect
import os
from functools import partial, wraps
from typing import Any, Callable, Dict, Optional, Type

import pandas as pd
import sarus_data_spec.typing as st
import yaml
from sarus_data_spec.config import WHITELISTED_TRANSFORMS
from sarus_data_spec.context import global_context
from sarus_data_spec.scalar import model
from sarus_data_spec.transform import external
from sarus_data_spec.variant_constraint import variant_constraint

from .context.typing import LocalSDKContext
from .typing import DataSpecVariant, DataSpecWrapper

config_file = os.path.join(os.path.dirname(__file__), "config.yaml")
with open(config_file) as f:
    config = yaml.load(f.read(), Loader=yaml.Loader)


def module_config(module_name: str) -> Optional[Dict[str, Any]]:
    """Fetch the module's configuration from the config dict."""
    keys = module_name.split(".")
    module_conf = config
    for key in keys:
        if module_conf is None:
            return
        module_conf = module_conf.get(key)
    return module_conf


def eval(x: Any) -> st.DataSpecValue:
    if isinstance(x, DataSpecWrapper):
        return x.__sarus_eval__()
    elif isinstance(x, list):
        return [eval(x_) for x_ in x]
    elif isinstance(x, tuple):
        return tuple([eval(x_) for x_ in x])
    elif isinstance(x, dict):
        return {eval(k): eval(v) for k, v in x.items()}
    else:
        return x


def eval_policy(x: Any) -> Optional[str]:
    """The alternative dataspec's privacy policy."""
    if isinstance(x, DataSpecWrapper):
        return x.__eval_policy__()
    else:
        return None


_registered_methods = []
_registered_functions = []


class register_method:
    """This decorator method allows to register methods declared in classes.

    It uses this behavior since Python 3.6
    https://docs.python.org/3/reference/datamodel.html#object.__set_name__
    """

    def __init__(self, method: Callable, code_name: str) -> None:
        self.method = method
        self.code_name = code_name

    def __set_name__(self, owner: Type, name: str) -> None:
        global _registered_methods
        _registered_methods.append(
            (owner.__module__, owner.__name__, name, self.code_name)
        )
        setattr(owner, name, self.method)


def register_ops():
    """Monkey-patching standard libraries to have Sarus functions.

    This functions is intended to be called in a Sarus module. The module's
    local variables will be modified dynamically (monkey patching) to replace
    some functions or methods by Sarus equivalent operations.

    Technically, we get the previous frame's (the module where the function is
    called) locals mapping and update it.

    The modified methods and functions are listed in the `sarus/config.yaml`
    file.
    """
    previous_frame = inspect.currentframe().f_back
    local_vars = previous_frame.f_locals
    module_name = local_vars["__name__"]
    module_conf = module_config(module_name)
    if module_conf is None:
        return

    # Registering module functions
    global _registered_functions
    functions = module_conf.get("sarus_functions", {})
    for fn_name, fn_code_name in functions.items():
        local_vars[fn_name] = create_op(fn_code_name)
        _registered_functions.append((module_name, fn_name, fn_code_name))

    # Registering explicit evaluation functions
    explicit_eval_fns = module_conf.get("explicit_eval", [])
    for fn_name in explicit_eval_fns:
        fn_obj = local_vars[fn_name]
        local_vars[fn_name] = explicit_sarus_eval(fn_obj)

    # Registering classes methods
    global _registered_methods
    classes = module_conf.get("classes", {})
    for class_name, methods in classes.items():
        class_obj = local_vars[class_name]
        for mth_name, mth_code_name in methods.items():
            _registered_methods.append(
                (module_name, class_name, mth_name, mth_code_name)
            )
            setattr(class_obj, mth_name, create_op(mth_code_name))


def _sarus_op(
    code_name: str,
    inplace: bool = False,
    register: bool = False,
    is_property: bool = False,
):
    """Parametrized decorator to register a Sarus external op."""

    def parametrized_wrapper(ops_fn):
        @wraps(ops_fn)
        def wrapper_fn(*args, **kwargs):
            """Some arguments are instances of DataSpecWrapper and others are
            just Python object. This wrapper registers a new dataspec."""
            py_args = {
                i: eval(arg)
                for i, arg in enumerate(args)
                if not isinstance(arg, DataSpecWrapper)
            }
            ds_args_pos = [
                i
                for i, arg in enumerate(args)
                if isinstance(arg, DataSpecWrapper)
            ]
            ds_args = [
                arg.dataspec(DataSpecVariant.USER_DEFINED)
                for arg in args
                if isinstance(arg, DataSpecWrapper)
            ]
            py_kwargs = {
                eval(name): eval(arg)
                for name, arg in kwargs.items()
                if not isinstance(arg, DataSpecWrapper)
            }
            ds_kwargs = {
                name: arg.dataspec(DataSpecVariant.USER_DEFINED)
                for name, arg in kwargs.items()
                if isinstance(arg, DataSpecWrapper)
            }
            transform = external(
                id=code_name,
                py_args=py_args,
                py_kwargs=py_kwargs,
                ds_args_pos=ds_args_pos,
            )
            new_dataspec = transform(*ds_args, **ds_kwargs)
            context: LocalSDKContext = global_context()

            new_dataspec_wrapper = context.wrapper_factory().create(
                new_dataspec
            )

            if inplace:
                self: DataSpecWrapper = args[0]  # TODO check semantic
                self._set_dataspec(new_dataspec)

            return new_dataspec_wrapper

        if is_property:
            wrapper_fn = property(wrapper_fn)

        if register:
            wrapper_fn = register_method(wrapper_fn, code_name)

        return wrapper_fn

    return parametrized_wrapper


sarus_method = partial(_sarus_op, register=True, is_property=False)
sarus_property = partial(_sarus_op, register=True, is_property=True)


def sarus_model(code):
    """Decorator to initialize models."""

    def parametrized_wrapper(ops_fn):
        @wraps(ops_fn)
        def wrapper_fn(self, *args, **kwargs):
            dataspec = model(model_class=code, *args, **kwargs)
            variant_constraint(st.ConstraintKind.PUBLIC, dataspec)
            self._set_dataspec(dataspec)

        wrapper_fn = register_method(wrapper_fn, code)

        return wrapper_fn

    return parametrized_wrapper


def create_op(code_name: str, inplace: bool = False) -> Callable:
    """Create an op function without decorating a function."""

    @_sarus_op(code_name=code_name, inplace=inplace)
    def dummy_fn(*args, **kwargs):
        ...

    return dummy_fn


@_sarus_op(code_name="std.LEN")
def length(__o: object):
    ...


def explicit_sarus_eval(func):
    """Decorator to explicitly collect Dataspec's values before calling."""

    @wraps(func)
    def wrapped_func(*args, **kwargs):
        args = [eval(arg) for arg in args]
        kwargs = {key: eval(val) for key, val in kwargs.items()}
        return func(*args, **kwargs)

    return wrapped_func


def init_wrapped(wrapper_class):
    """Define the constructor to return a wrapped instance."""
    assert issubclass(wrapper_class, DataSpecWrapper)

    def __new__(cls, *args, **kwargs):
        return wrapper_class.__wraps__(*args, **kwargs)

    wrapper_class.__new__ = staticmethod(__new__)

    return wrapper_class


def generate_op_list():
    """Generate the list of registered operations in a Markdown file.

    NB: This does not list DataspecWrappers without any operations declared.
    """
    global _registered_functions, _registered_methods

    methods = pd.DataFrame.from_records(_registered_methods)
    methods.columns = ["module", "class", "method", "code"]

    functions = pd.DataFrame.from_records(_registered_functions)
    functions.columns = ["module", "function", "code"]

    all_items = methods.append(functions, ignore_index=True)

    lines, whitelisted_lines = [], []
    for mod_name, mod_df in all_items.groupby(by="module"):
        has_whitelisted = False
        mod_whitelisted_lines = []

        lines.append(f"\n# {mod_name}")
        mod_whitelisted_lines.append(f"# {mod_name}")

        fns = (
            mod_df.loc[:, ["function", "code"]]
            .dropna()
            .sort_values(by="function")
        )
        mask = fns.code.apply(lambda x: x in WHITELISTED_TRANSFORMS)
        whitelisted_fns = fns[mask]
        if len(fns) > 0:
            lines.append("\n## Functions")
            lines += list(map(lambda x: f"- `{x}`", fns.function))
        if len(whitelisted_fns) > 0:
            has_whitelisted = True
            mod_whitelisted_lines.append("\n## Functions")
            mod_whitelisted_lines += list(
                map(lambda x: f"- `{x}`", whitelisted_fns.function)
            )

        for class_name, class_df in mod_df.groupby("class"):
            if class_name == "DataSpecWrapper":
                class_name = "Generic Operations"
            lines.append(f"\n## {class_name}")
            methods = (
                class_df.loc[:, ["method", "code"]]
                .dropna()
                .sort_values(by="method")
            )
            mask = methods.code.apply(lambda x: x in WHITELISTED_TRANSFORMS)
            lines += list(map(lambda x: f"- `{x}`", methods.method))

            whitelisted_methods = methods[mask]
            if len(whitelisted_methods) > 0:
                has_whitelisted = True
                mod_whitelisted_lines.append(f"\n## {class_name}")
                mod_whitelisted_lines += list(
                    map(lambda x: f"- `{x}`", whitelisted_methods.method)
                )

        if has_whitelisted:
            whitelisted_lines += mod_whitelisted_lines

    ops = "\n".join(lines)
    whitelisted_ops = "\n".join(whitelisted_lines)

    op_file = os.path.join(os.path.dirname(__file__), "op_list.md")
    with open(op_file, "w") as f:
        f.write(ops)

    whitelisted_op_file = os.path.join(
        os.path.dirname(__file__), "whitelisted_op_list.md"
    )
    with open(whitelisted_op_file, "w") as f:
        f.write(whitelisted_ops)


if __name__ == "__main__":
    generate_op_list()



from typing import Optional, Protocol

import sarus_data_spec.typing as st
from sarus_data_spec.manager.typing import Manager


class LocalSDKManager(Manager, Protocol):
    def mock(self, dataspec: st.DataSpec) -> st.DataSpec:
        """Returns a mock version of a DataSpec.

        This is a kind of compilation used by this manager to infer the DataSpec type and Python type of an external transform's result.
        """
        ...

    def register_source_mock(self, dataspec: st.DataSpec) -> st.DataSpec:
        """Define what a mock is for source or fetched DataSpecs."""
        ...

    def python_type(self, dataspec: st.DataSpec) -> Optional[str]:
        ...
